/*
 *  linux/drivers/i2c/i2c-armadillo9.c
 *
 *  Copyright (C) 2004-2007 Atmark Techno, Inc.
 *
 * For Armadillo-220, 230, and 240, we use
 * - PFDR 0x20 as clock line, and
 * - PDDR 0x80 as data line.
 *
 * For Armadillo-9, we use
 * - PBDR 0x20 as clock line, and
 * - PBDR 0x10 as data line.
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/time.h>
#include <linux/miscdevice.h>
#include <linux/rtc.h>
#include <linux/i2c.h>
#include <linux/i2c-algo-bit.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/rtc.h>

#include "../chips/i2c-s3531a.h"
#include "../chips/i2c-at24cxx.h"

#define AUTHOR "Atmark Techno, Inc."
#define DESCRIPTION "i2c Armadillo-9 driver"
#define LICENSE "GPL v2"

#define I2C_HW_B_ARMADILLO2X0_1	(0x16)
#define I2C_HW_B_ARMADILLO2X0_2	(0x17)

#define CHRDEV  "i2c-armadillo9"
#define VERSION CHRDEV ": " DESCRIPTION ", (C) 2004-2007 " AUTHOR

#define LOWLEVEL_UDELAY		(20)
#define LOWLEVEL_MDELAY		(20)
#define LOWLEVEL_TIMEOUT	(100)

struct slave_info {
	int (*init)(void);
	void (*exit)(void);
	int (*client_reg_ext)(void);
	int (*client_unreg_ext)(void);
	int addr;
	struct i2c_client *client;
};

struct gpio_port_reg {
	u32 ddr;
	u32 dr;
};

struct gpio_scl_sda {
	struct gpio_port_reg *port;
	u8 mask;
};

struct a9_i2c_lowlevel {
	struct gpio_scl_sda scl;
	struct gpio_scl_sda sda;
	u32 id;
};

static struct gpio_port_reg portb = {
	.ddr	= GPIO_PBDDR,
	.dr	= GPIO_PBDR,
};

#if defined(CONFIG_I2C_ARMADILLO2X0_EXT)
static struct gpio_port_reg portd = {
	.ddr	= GPIO_PDDDR,
	.dr	= GPIO_PDDR,
};

static struct gpio_port_reg portf = {
	.ddr	= GPIO_PFDDR,
	.dr	= GPIO_PFDR,
};
#endif

static struct a9_i2c_lowlevel lowlevel_priv[] = {
  {
	.scl	= {
		.port	= &portb,
		.mask	= 1 << 5,
	},
	.sda	= {
		.port	= &portb,
		.mask	= 1 << 4,
	},
	.id	= I2C_HW_B_ARMADILLO2X0_1,
  },
#if defined(CONFIG_I2C_ARMADILLO2X0_EXT)
  {
	.scl	= {
		.port	= &portf,
		.mask	= 1 << 5,
	},
	.sda	= {
		.port	= &portd,
		.mask	= 1 << 7,
	},
	.id	= I2C_HW_B_ARMADILLO2X0_2,
  },
#endif
};
#if defined(CONFIG_I2C_S3531A)
static int rtc_device_init(void);
static void rtc_device_exit(void);
static int rtc_client_register_extra(void);
#endif

#if defined(CONFIG_I2C_EEPROM)
static int eeprom_device_init(void);
static void eeprom_device_exit(void);
#endif

/*****************************************************************************
 * Slave
 *****************************************************************************/
static struct slave_info slave_info[] = {
#if defined(CONFIG_I2C_S3531A)
  {
    .init = rtc_device_init,
    .exit = rtc_device_exit,
    .client_reg_ext = rtc_client_register_extra,
    .addr = I2C_S3531A_ADDRESS,
  },
#endif
#if defined(CONFIG_I2C_EEPROM)
  {
    .init = eeprom_device_init,
    .exit = eeprom_device_exit,
    .addr = I2C_EEPROM_ADDRESS,
  },
#endif
};

/*****************************************************************************
 * RTC
 *****************************************************************************/
#if defined(CONFIG_I2C_S3531A)
static inline int rtc_command(int cmd, void *data)
{
  int ret = -EIO;
  int i;

  for (i=0; i<ARRAY_SIZE(slave_info); i++)
    if ((slave_info[i].addr == I2C_S3531A_ADDRESS) && slave_info[i].client)
      ret = slave_info[i].client->driver->command(slave_info[i].client,
						  cmd,
						  data);
  
  return ret;
}

static inline int rtc_sync(void)
{
  return rtc_command(RTC_SYNC, NULL);
}

static inline int get_rtc_time(struct rtc_time *rtctm)
{
  struct s3531a_time rawtm;
  int ret;
  
  memset(rtctm, 0, sizeof(struct rtc_time));
  
  ret = rtc_command(RTC_GETDATETIME, &rawtm);
  if (!ret) {
    rtctm->tm_year = rawtm.year + ((rawtm.year <= 69) ? 100 : 0);
    rtctm->tm_mon  = rawtm.mon - 1;
    rtctm->tm_mday = rawtm.mday;
    rtctm->tm_wday = rawtm.wday;
    rtctm->tm_hour = rawtm.hour;
    rtctm->tm_min  = rawtm.min;
    rtctm->tm_sec  = rawtm.sec;
  }
  
  return ret;
}

static inline int set_rtc_time(struct rtc_time *rtctm)
{
  struct s3531a_time rawtm;
  
  rawtm.year = rtctm->tm_year - ((rtctm->tm_year < 100) ? 0 : 100);
  rawtm.mon  = rtctm->tm_mon + 1;
  rawtm.mday = rtctm->tm_mday;
  rawtm.wday = rtctm->tm_wday;
  rawtm.hour = rtctm->tm_hour;
  rawtm.min  = rtctm->tm_min;
  rawtm.sec  = rtctm->tm_sec;
  
  return rtc_command(RTC_SETDATETIME, &rawtm);
}

static int rtc_ioctl(struct inode *inode, struct file *file,
		     unsigned int cmd, unsigned long arg)
{
  struct rtc_time rtctm;
  int ret;
  
  switch (cmd) {
  case RTC_ALM_READ:
    break;
    
  case RTC_ALM_SET:
    break;
    
  case RTC_RD_TIME:
    ret = get_rtc_time(&rtctm);
    if (ret) {
      return ret;
    }
    return copy_to_user((void *)arg, &rtctm, sizeof(rtctm)) ? -EFAULT : 0;
    
  case RTC_SET_TIME:
    if (copy_from_user(&rtctm, (struct rtc_time*)arg, sizeof(rtctm))) {
      return -EFAULT;
    }
    return set_rtc_time(&rtctm);
  }
  
  return -EINVAL;
}

static struct file_operations rtc_fops = {
  ioctl: rtc_ioctl,
};

static struct miscdevice rtc_dev = {
  minor: RTC_MINOR,
  name:	 "rtc",
  fops:	 &rtc_fops,
};

static int
rtc_client_register_extra(void)
{
  struct rtc_time now;
  struct timespec time;

  if (rtc_sync()) {
    printk (KERN_WARNING CHRDEV ": RTC stopped.\n");
    return -EIO;
  }

  if (get_rtc_time(&now)) {
    printk (KERN_WARNING CHRDEV ": RTC read failed.\n");
    return -EIO;
  }

  time.tv_nsec = 0;
  rtc_tm_to_time(&now, &time.tv_sec);
  do_settimeofday(&time);

  return 0;
}

static int
rtc_device_init(void)
{
  int ret;

  ret = misc_register(&rtc_dev);
  if (ret < 0) {
    printk (KERN_WARNING CHRDEV ": Unable to register %s driver.\n",
	    rtc_dev.name);
    return ret;
  }
  return 0;
}

static void
rtc_device_exit(void)
{
  misc_deregister(&rtc_dev);
}
#endif

/*****************************************************************************
 * EEPROM
 *****************************************************************************/
#if defined(CONFIG_I2C_EEPROM)
int eeprom_ioctl(struct inode *inode, struct file *file,
		 unsigned int cmd, unsigned long arg)
{
  int ret = -EIO;
  int i;
  
  for (i=0; i<ARRAY_SIZE(slave_info); i++)
    if (slave_info[i].addr == I2C_EEPROM_ADDRESS && slave_info[i].client)
      ret = slave_info[i].client->driver->command(slave_info[i].client,
						  cmd,
						  (void *)arg);

  return ret;
}

static struct file_operations eeprom_fops = {
  ioctl: eeprom_ioctl,
};

static int
eeprom_device_init(void)
{
  int ret = 0;

  ret = register_chrdev(EEPROM_MAJOR, "eeprom", &eeprom_fops);
  if (ret < 0) {
    printk(KERN_WARNING CHRDEV ": Unable to register EEPROM driver.\n");
    return ret;
  }
  return 0;
}

static void
eeprom_device_exit(void)
{
  unregister_chrdev(EEPROM_MAJOR, "eeprom");
}
#endif

/*****************************************************************************
 * Bus Master
 *****************************************************************************/
static void lowlevel_setscl(void *data, int state)
{
  struct a9_i2c_lowlevel *priv = (struct a9_i2c_lowlevel *)data;
  struct gpio_scl_sda *scl = &priv->scl;

  if (state)
    outl(inl(scl->port->dr) | scl->mask, scl->port->dr);
  else
    outl(inl(scl->port->dr) & ~scl->mask, scl->port->dr);
  outl(inl(scl->port->ddr) | scl->mask, scl->port->ddr);
}

static void lowlevel_setsda(void *data, int state)
{
  struct a9_i2c_lowlevel *priv = (struct a9_i2c_lowlevel *)data;
  struct gpio_scl_sda *sda = &priv->sda;

  if (state)
    outl(inl(sda->port->ddr) & ~sda->mask, sda->port->ddr);
  else {
    outl(inl(sda->port->dr) & ~sda->mask, sda->port->dr);
    outl(inl(sda->port->ddr) | sda->mask, sda->port->ddr);
  }
}

static int lowlevel_getscl(void *data)
{
  struct a9_i2c_lowlevel *priv = (struct a9_i2c_lowlevel *)data;
  struct gpio_scl_sda *scl = &priv->scl;

  return (inl(scl->port->dr) & scl->mask) ? 1 : 0;
}

static int lowlevel_getsda(void *data)
{
  struct a9_i2c_lowlevel *priv = (struct a9_i2c_lowlevel *)data;
  struct gpio_scl_sda *sda = &priv->sda;

  return (inl(sda->port->dr) & sda->mask) ? 1 : 0;
}

static int lowlevel_initialize(void *data)
{
  struct a9_i2c_lowlevel *priv = (struct a9_i2c_lowlevel *)data;

  /* Initialize SCL/SDA */
  lowlevel_setscl(priv, 1);
  lowlevel_setsda(priv, 1);

  return 0;
}

static struct i2c_algo_bit_data armadillo2x0_data[] = {
  {
    .setsda	= lowlevel_setsda,
    .setscl	= lowlevel_setscl,
    .getsda	= lowlevel_getsda,
    .getscl	= lowlevel_getscl,
    .udelay	= LOWLEVEL_UDELAY,
    .mdelay	= LOWLEVEL_MDELAY,
    .timeout	= LOWLEVEL_TIMEOUT,
    .data	= &lowlevel_priv[0],
  },
#if defined(CONFIG_I2C_ARMADILLO2X0_EXT)
  {
    .setsda	= lowlevel_setsda,
    .setscl	= lowlevel_setscl,
    .getsda	= lowlevel_getsda,
    .getscl	= lowlevel_getscl,
    .udelay	= LOWLEVEL_UDELAY,
    .mdelay	= LOWLEVEL_MDELAY,
    .timeout	= LOWLEVEL_TIMEOUT,
    .data	= &lowlevel_priv[1],
  },
#endif
};

static int armadillo2x0_client_reg(struct i2c_client *client)
{
  int ret = -EINVAL;
  int i;

  for (i=0; i<ARRAY_SIZE(slave_info); i++)
    if (client->addr == slave_info[i].addr) {
      slave_info[i].client = client;
      ret = 0;
      if (slave_info[i].client_reg_ext) {
	ret = slave_info[i].client_reg_ext();
	if (ret < 0)
	  slave_info[i].client = NULL;
      }
    }

  return ret;
}

static int armadillo2x0_client_unreg(struct i2c_client *client)
{
  int ret = 0;
  int i;

  for (i=0; i<ARRAY_SIZE(slave_info); i++)
    if (client->addr == slave_info[i].addr) {
      if (slave_info[i].client && slave_info[i].client_unreg_ext)
	ret = slave_info[i].client_unreg_ext();
      slave_info[i].client = NULL;
    }
  
  return ret;
}

static struct i2c_adapter armadillo2x0_ops[] = {
  {
    .name		= "I2C1",
    .id			= I2C_HW_B_ARMADILLO2X0_1,
    .algo_data		= &armadillo2x0_data[0],
    .client_register	= armadillo2x0_client_reg,
    .client_unregister	= armadillo2x0_client_unreg,
  },
#if defined(CONFIG_I2C_ARMADILLO2X0_EXT)
  {
    .name		= "I2C2",
    .id			= I2C_HW_B_ARMADILLO2X0_2,
    .algo_data		= &armadillo2x0_data[1],
    .client_register	= armadillo2x0_client_reg,
    .client_unregister	= armadillo2x0_client_unreg,
  },
#endif
};

static int __init i2c_armadillo9_init(void)
{
  int i;
  int ret;

  for (i=0; i<ARRAY_SIZE(armadillo2x0_ops);i++) {
    lowlevel_initialize(&lowlevel_priv[i]);

    ret = i2c_bit_add_bus(&armadillo2x0_ops[i]);
    if (ret < 0) {
      printk (KERN_WARNING CHRDEV ": Unable to add %s bus.\n",
	      armadillo2x0_ops[0].name);
      goto master_init_err;
    }
  }

  for (i=0; i<ARRAY_SIZE(slave_info); i++) {
    if (slave_info[i].init)
      slave_info[i].init();
  }
  
  printk (KERN_INFO VERSION "\n");

  return 0;

 master_init_err:
  return ret;
}

static void __exit i2c_armadillo9_exit(void)
{
  int i;

  for (i=0; i<ARRAY_SIZE(slave_info); i++)
    if (slave_info[i].exit)
      slave_info[i].exit();

  for (i=0; i<ARRAY_SIZE(armadillo2x0_ops);i++)
    i2c_bit_del_bus(&armadillo2x0_ops[i]);
}

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESCRIPTION);
MODULE_LICENSE(LICENSE);

module_init(i2c_armadillo9_init);
module_exit(i2c_armadillo9_exit);
