/*
 * linux/drivers/ide/arm/ide-dma-ep93xx.c
 * Copyright (C) 2007 Atmark Techno, Inc
 *
 * Adapted from original the ide-dma-ep93xx.c, which is:
 *   Copyright (C) 2005 Cirrus Logic
 *
 * and which in turn was adapted from ide-dma.c
 *   Copyright (c) 1999-2000 Andre Hedrick <andre@linux-ide.org>
 *
 * and also pmac.c, which is:
 *   Copyright (C) 1998-2003 Paul Mackerras & Ben. Herrenschmidt
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 */

#include <linux/ide.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/ide.h>
#include <asm/irq.h>
#include <asm/arch/ide.h>
#include <asm/arch/dma.h>
#include <asm/hardware.h>

#include <linux/cdrom.h>
#include <linux/seq_file.h>
#include <linux/device.h>
#include <linux/dma-mapping.h>

#undef DEBUG
//#define DEBUG 1
#ifdef DEBUG
#  define DPRINTK(fmt, args...)	printk("%s: " fmt, __FUNCTION__ , ## args)
#else
#  define DPRINTK(fmt, args...)
#endif

static unsigned int *g_dma_table_base;

#define RETRIES_BEFORE_PIO_TRANSFER	(ERROR_RESET - 1)
#define FAILURES_BEFORE_DMA_BACKOFF	50

/*
 * Global to keep track of the number of entries in the dma buffer
 * table for each transfer.
 */
static int g_prd_count;
static unsigned int g_prd_total;
static unsigned int g_prd_returned;

/*
 * Set during the DMA callback function to indicate that a ep93xx rev e0
 * related DMA error occurred 
 */
static unsigned int g_bad = 0;

/* The particulars of the most recent command that was executed. */
static unsigned long g_cmd 		= 0;
static ide_drive_t *g_drive 		= NULL;
static unsigned long g_sector 		= 0;
static unsigned long g_nr_sectors 	= 0;
static unsigned long g_cur_retries 	= 0;

/*
 * The count of consective DMA request errors.
 * { hda reads, hda writes, hdb reads, hdb writes }
 */
static unsigned long g_errors[4] = { 0, 0, 0, 0 };

/*
 * A garbage word that is used to satify additional DMA requests from the
 * device.
 */
static unsigned long g_garbage;

/*
 * ep93xx_ide_dma_intr() - handler for disk read/write DMA interrupts
 */
static ide_startstop_t ep93xx_ide_dma_intr(ide_drive_t *drive)
{
	byte stat, dma_stat;

	DPRINTK("%s\n", drive->name);

	dma_stat = HWIF(drive)->ide_dma_end(drive);
	stat = HWIF(drive)->INB(IDE_STATUS_REG);
	if (OK_STAT(stat,DRIVE_READY,drive->bad_wstat|DRQ_STAT)) {
		if (!dma_stat) {
			struct request *rq = HWGROUP(drive)->rq;
			ide_driver_t *drv = *(ide_driver_t **)rq->rq_disk->private_data;
			drv->end_request(drive, 1, rq->nr_sectors);
			return ide_stopped;
		}
		printk(KERN_ERR "%s: dma_intr: bad DMA status (dma_stat=%x)\n",
		       drive->name, dma_stat);
   	}

	return ide_error(drive, __FUNCTION__, stat);
}

/*
 * ide_build_sglist() - builds a table of buffers to be used by the dma.
 */
static int ide_build_sglist(ide_drive_t *drive, struct request *rq)
{
	ide_hwif_t *hwif = HWIF(drive);
	struct scatterlist *sg = hwif->sg_table;

	DPRINTK("%s: sg_table 0x%08lx\n", drive->name,
		(unsigned long)hwif->sg_table);

	if ((rq->flags & REQ_DRIVE_TASKFILE) && rq->nr_sectors > 256)
		BUG();

	ide_map_sg(drive, rq);

	if (rq_data_dir(rq) == READ)
		hwif->sg_dma_direction = PCI_DMA_FROMDEVICE;
	else
		hwif->sg_dma_direction = PCI_DMA_TODEVICE;

     	return dma_map_sg(NULL, sg, hwif->sg_nents, hwif->sg_dma_direction);
}

/*
 * ide_build_dmatable - prepares the dma table for a request.
 *
 * See ide_build_dmatable() in ide-dma.c for comments.
 */
static int ide_build_dmatable(ide_drive_t *drive,struct request *rq)
{       
	ide_hwif_t *hwif  = HWIF(drive);
	/* todo: why can't we just use dmatable_cpu? is it get damaged somewhere? */
 	unsigned int *table = HWIF(drive)->dmatable_cpu = g_dma_table_base;
	unsigned int count = 0;
	int i;
	struct scatterlist *sg;

	DPRINTK("%s\n", drive->name);

	hwif->sg_nents = i = ide_build_sglist(drive, rq);

	if (!i)
		return 0;

	sg = hwif->sg_table;
	while (i) {
		u32 cur_addr;
		u32 cur_len;

		cur_addr = sg_dma_address(sg);
		cur_len = sg_dma_len(sg);

		/*
		 * Fill in the dma table, without crossing any 64kB boundaries.
		 * Most hardware requires 16-bit alignment of all blocks,
		 * but the trm290 requires 32-bit alignment.
		 */
		while (cur_len) {
			if (count++ >= PRD_ENTRIES) {
				printk(KERN_ERR"%s: DMA table too small\n", drive->name);
				goto use_pio_instead;
			} else {
				u32 xcount, bcount = 0x10000 - (cur_addr & 0xffff);

				if (bcount > cur_len)
					bcount = cur_len;
				*table++ = cur_addr;
				xcount = bcount & 0xffff;
				if (xcount == 0x0000) {
					/*
					 * Most chipsets correctly interpret a
					 * length of 0x0000 as 64KB, but at
					 * least one (e.g. CS5530)
					 * misinterprets it as zero (!).  So
					 * here we break the 64KB entry into
					 * two 32KB entries instead.
					 */
					if (count++ >= PRD_ENTRIES) {
						printk(KERN_ERR"%s: DMA table too small\n",
						       drive->name);
						goto use_pio_instead;
					}
					*table++ = 0x8000;
					*table++ = cur_addr + 0x8000;
					xcount = 0x8000;
				}
				*table++ = xcount;
				cur_addr += bcount;
				cur_len -= bcount;
			}
		}
		sg++;
		i--;
	}

	if (count) {
		*--table |= 0x80000000;
		return count;
	}
	printk(KERN_ERR"%s: empty DMA table?\n", drive->name);

use_pio_instead:

	dma_unmap_sg(NULL, hwif->sg_table, hwif->sg_nents,
		hwif->sg_dma_direction);

	return 0;
}

/*
 * ep93xx_ide_callback() - cleanup after dma transfer
 *
 * Registered with the ep93xx dma driver and called at the end of the dma
 * interrupt handler, this function processes the dma buffers.
 */
static void ep93xx_ide_callback(ep93xx_dma_int_t dma_int, 
		ep93xx_dma_dev_t device, unsigned int user_data)
{
	ide_drive_t *drive = (ide_drive_t *)user_data;
	struct request *rq = HWGROUP(drive)->rq;
	ide_hwif_t *hwif = HWIF(drive);
	unsigned int temp;

	DPRINTK("%s\n", drive->name);

	/*
	 * Retrieve from the dma interface as many used buffers as are
	 * available.
	 */
	while (ep93xx_dma_remove_buffer(hwif->hw.dma, &temp) == 0)
		g_prd_returned++;

	/* add any remaining buffers */
	while (g_prd_count) {
		if (ep93xx_dma_add_buffer(hwif->hw.dma,
					  rq_data_dir(rq) == READ ?
					  hwif->dma_base :
					  hwif->dmatable_cpu[0],
					  rq_data_dir(rq) == READ ?
					  hwif->dmatable_cpu[0] :
					  hwif->dma_base,
					  hwif->dmatable_cpu[1], 0,
					  g_prd_count) != 0)
			break;
		hwif->dmatable_cpu += 2;
		g_prd_count--;
	};

	/*
	 * check if all of the buffers have been returned, but the device
	 * still wants more.
	 */
   	if ((g_prd_returned >= g_prd_total) && (inl(IDECR) & IDECtrl_DMARQ)) {
		g_bad = 1;
		printk(KERN_WARNING "%s: DMA transfer error detected: "
			"transfer complete but device wants more\n",
			drive->name);
		if (hwif->sg_dma_direction == PCI_DMA_FROMDEVICE)
			ep93xx_dma_add_buffer(hwif->hw.dma,
					      hwif->dma_base,
					      virt_to_dma
					      (NULL, &g_garbage),
					      4, 0, 0);
		else
		     	ep93xx_dma_add_buffer(hwif->hw.dma,
		     			      virt_to_dma
					      (NULL, &g_garbage),
					      hwif->dma_base,
					      4, 0, 0);
	}

	DPRINTK("	return from callback\n");
}

/*
 * ep93xx_dma_timer_expiry() - handle a DMA timeout
 * 
 * See dma_timer_expiry() in ide-dma.c for comments.
 */
static int ep93xx_idedma_timer_expiry(ide_drive_t *drive)
{
	ide_hwif_t *hwif = HWIF(drive);
	u8 dev_stat	 = hwif->INB(IDE_ALTSTATUS_REG);
	u8 irq_stat	 = inl(IDECR) & IDECtrl_INTRQ;
	
	DPRINTK("%s\n", drive->name);
	printk(KERN_WARNING " \n %s: dma_timer_expiry: dev status == "
		"0x%02x,irq= %d\n", drive->name, dev_stat, irq_stat);

	/*
	 * Clear the expiry handler in case we decide to wait more,
	 * next time timer expires it is an error
	 */
	HWGROUP(drive)->expiry = NULL;

	/* If the interrupt is asserted, call the handler. */
	if (irq_stat)
		HWGROUP(drive)->handler(drive);

	/*
	 * Check if the busy bit or the drq bit is set, indicating that
	 * a dma transfer is still active, or the IDE interrupt is asserted.
	 */
	if ((dev_stat & 0x80) || (dev_stat & 0x08) || irq_stat)
		return WAIT_CMD;

	/*
	 * the device is not busy and the interrupt is not asserted, so check
	 * if there's an error.
	 */
	if (dev_stat & 0x01) /* ERROR */
		return -1;

	DPRINTK("%s (end)\n", drive->name);

	return 0; /* Unknown status -- reset the bus */
}

/*
 * ep93xx_config_ide_device() - set the max DMA speed for the device
 *
 * Determines the best transfer speed supported, and then configures the ide
 * device at that speed.
 * On Rev D1/E0 the maximum DMA mode is 2. 
 * On Rev E1 and later the maximum UDMA mode is 4.
 * At the moment, we're only allowing up to mode 3.
 */
static unsigned int ep93xx_config_ide_device(ide_drive_t *drive)
{
	unsigned int   ulChipID;
	byte transfer = 0;

	DPRINTK("%s\n", drive->name);

 	ulChipID = inl(SYSCON_CHIPID);
	transfer = ide_dma_speed(drive, 2);
	if (transfer >= XFER_UDMA_3) {
		if (((ulChipID &  SYSCON_CHIPID_REV_MASK)
		    >> SYSCON_CHIPID_REV_SHIFT) < 0x6) {
			transfer = XFER_UDMA_2;
		} else
			transfer = XFER_UDMA_3;
	}

	/* Configure the drive. */
	if (ide_config_drive_speed(drive, transfer)) {
		HWIF(drive)->ide_dma_off_quietly(drive);
		return 1;
	}

	return 0;
}

/*
 * ep93xx_rwproc() - initialize the ep93xx IDE controller
 *
 * Initializes the ep93xx IDE controller interface with the transfer type,
 * transfer mode, and transfer direction.
 */
static void ep93xx_rwproc(ide_drive_t *drive, int action)
{
	int speed;

	DPRINTK("%s\n", drive->name);

	/* Insure that neither device is selected. */
	outl(inl(IDECR) | (IDECtrl_CS0n | IDECtrl_CS1n), IDECR);

	/* Configure the IDE controller for the specified transfer mode. */
	switch (drive->current_speed) {

	case XFER_UDMA_0:
	case XFER_UDMA_1:
	case XFER_UDMA_2:
	case XFER_UDMA_3:
	case XFER_UDMA_4:

		speed = drive->current_speed;
		speed -= XFER_UDMA_0;
		if(speed >= 4)
			speed = 3; /* max udma 3 */
		outl(((speed & 7) << IDECfg_MODE_SHIFT) | IDECfg_UDMA |
			IDECfg_IDEEN, IDECFG);

		outl(action ? IDEUDMAOp_RWOP : 0, IDEUDMAOP);
		outl(inl(IDEUDMAOP) | IDEUDMAOp_UEN, IDEUDMAOP);
		break;

	default:
		break;
	}
}

/*
 * config_drive_for_dma - attempt to activate IDE DMA
 *
 * Enable DMA on any drive that has UltraDMA (mode 0/1/2/3/4/5/6) enabled.
 * ep93xx supports up to udma2. 
 */
static int config_drive_for_dma(ide_drive_t *drive)
{
	struct hd_driveid *id = drive->id;
	ide_hwif_t *hwif = HWIF(drive);

	DPRINTK("%s\n", drive->name);

	if ((id->capability & 1) &&
		hwif->autodma &&
		!__ide_dma_bad_drive(drive) &&
		(id->field_valid & 4) &&
		(id->dma_ultra & 0x7f)) {

		return hwif->ide_dma_on(drive);
	}

	return hwif->ide_dma_off_quietly(drive);
}

/*
 * ep93xx_ide_dma_check() - check for and enable dma
 *
 * Determines if the device supports dma transfers,
 * and enables dma if it does.
 */
static int ep93xx_ide_dma_check(ide_drive_t *drive)
{
	DPRINTK("%s\n", drive->name);

	config_drive_for_dma(drive);

	return ep93xx_config_ide_device(drive);
}

/* 
 * ep93xx_ide_dma_host_off() - disables dma on the host.
 * 
 * No need for this on ep93xx??
 */
static int ep93xx_ide_dma_host_off(ide_drive_t *drive)
{
	DPRINTK("%s\n", drive->name);
	return 0;
}

/*
 * ep93xx_ide_dma_off_quietly() - disables dma for the device.
 */
static int ep93xx_ide_dma_off_quietly(ide_drive_t *drive)
{
	DPRINTK("%s\n", drive->name);

	drive->using_dma = 0;
	ide_toggle_bounce(drive, 0);

	return (HWIF(drive)->ide_dma_host_off(drive));
}

/*
 * ep93xx_ide_dma_on() - enables dma for the device.
 */
static int ep93xx_ide_dma_on(ide_drive_t *drive)
{
	DPRINTK("%s\n", drive->name);

	drive->using_dma = 1;
	ide_toggle_bounce(drive, 1);

	return HWIF(drive)->ide_dma_host_on(drive);
}

/*
 * ep93xx_ide_dma_host_on() - should enable DMA on a IDE controller
 *
 * Is acutally enabled via ep93xx_ide_dma_start()
 */
static int ep93xx_ide_dma_host_on(ide_drive_t *drive)
{
	DPRINTK("%s\n", drive->name);
	return 0;
}

/*
 * ep93xx_ide_dma_setup() - prepare for a DMA transfer
 *
 * Prepares the DMA scatter gather tables and DMA controller for
 * a IDE transfer.
 *
 * Returns 0 on success. If a PIO fallback is required then
 * 1 is returned. 
 */
static int ep93xx_ide_dma_setup(ide_drive_t *drive)
{
	struct request *rq = HWGROUP(drive)->rq;
	unsigned int flags, i;
	ide_hwif_t *hwif = HWIF(drive);

	DPRINTK("%s\n", drive->name);
	DPRINTK("    %ld, %ld\n", rq->sector, rq->nr_sectors);

	if (drive->waiting_for_dma) {
		DPRINTK("%s: dma already active \n", drive->name);
		return 1;
	}

	/* Compute the array index used for this request. */
	i = rq_data_dir(rq) == READ ? 0 : 1;
	i += drive->name[2] == 'a' ? 0 : 2;

	/*
	 * See if this request matches the most recent request, and that the
	 * most recent request ended in error.
	 */
	if (g_bad && (g_cmd == rq_data_dir(rq)) && (g_drive == drive) &&
	    (g_sector == rq->sector) && (g_nr_sectors == rq->nr_sectors)) {

		g_cur_retries++;
		if (g_cur_retries == RETRIES_BEFORE_PIO_TRANSFER) {
			printk(KERN_WARNING "%s: using pio for this request\n",
				drive->name);
			g_bad = 0;
			return 1;
		}

	} else {

		g_cur_retries = 0;
		g_cmd = rq_data_dir(rq);
		g_drive = drive;
		g_sector = HWGROUP(drive)->rq->sector;
		g_nr_sectors = HWGROUP(drive)->rq->nr_sectors;

	}

	/*
	 * Configure DMA M2M channel flags for a source/destination address
	 * hold, h/w initiated P2M/M2P transfer.
	 */
	flags = rq_data_dir(rq) == READ ?
	  (SOURCE_HOLD | TRANSFER_MODE_HW_P2M) :
	  (DESTINATION_HOLD | TRANSFER_MODE_HW_M2P);

	if (drive->current_speed & 0x20) {
		DPRINTK("%s: attempt to setup MDMA\n", drive->name);		
		return 1;
	} else {
		flags |= ((rq_data_dir(rq) == READ ?
			   WS_IDE_UDMA_READ : WS_IDE_UDMA_WRITE)
			  << WAIT_STATES_SHIFT);

		/* UDMA data register address. */
		hwif->dma_base = virt_to_dma(NULL,
		   (rq_data_dir(rq) == READ ? IDEUDMADATAIN : IDEUDMADATAOUT)
		   - IO_BASE_VIRT + IO_BASE_PHYS);
	}

	/* Configure the dma interface for this IDE operation. */
 	DPRINTK("\n\n start setup ep93xx_dma_config\n \n");
	if (ep93xx_dma_config(hwif->hw.dma, 0, flags, ep93xx_ide_callback,
			      (unsigned int)drive) != 0) {
		DPRINTK("%s: ep93xx_ide_dma_setup: ERROR- dma config failed",
				drive->name);
		return 1;
	}
	DPRINTK("\n\n finish setup ep93xx_dma_config\n\n");

	/* Build the table of dma-able buffers. */
	if (!(g_prd_count = ide_build_dmatable(drive,rq))) {
		DPRINTK("%s: ep93xx_ide_dma_setup: failed to build dma table",
			drive->name);
		return 1;
	}

	g_prd_total = g_prd_count;
	g_prd_returned = 0;

	/* Add buffers to the DMA interface from the dma_table. */
	do {
		if (ep93xx_dma_add_buffer(hwif->hw.dma,
					  rq_data_dir(rq) == READ ?
					  hwif->dma_base :
					  hwif->dmatable_cpu[0],
					  rq_data_dir(rq) == READ ?
					  hwif->dmatable_cpu[0] :
					  hwif->dma_base,
					  hwif->dmatable_cpu[1], 0,
					  g_prd_count) != 0)
			break;
		hwif->dmatable_cpu += 2;
		g_prd_count--;
	} while (g_prd_count);

	/* Indicate that we're waiting for dma. */
	drive->waiting_for_dma = 1;

	return 0;
}

/*
 * ep93xx_ide_dma_exec_cmd() - issues command to the drive
 */
static void ep93xx_ide_dma_exec_cmd(ide_drive_t *drive, u8 command)
{
	DPRINTK("%s: %s\n", drive->name, command == READ ? "read" : "write");

	ide_execute_command(drive, command, &ep93xx_ide_dma_intr, 2*WAIT_CMD,
			    &ep93xx_idedma_timer_expiry);

	DPRINTK("%s: %s (end)\n", drive->name,
		command == READ ? "read" : "write");
}

/*
 * ep93xx_ide_dma_start() - initiates a dma transfer.
 */
static void ep93xx_ide_dma_start(ide_drive_t *drive)
{
	ide_hwif_t *hwif = HWIF(drive);

	DPRINTK("%s\n", drive->name);

	/* the transfer is not bad yet. */
	g_bad = 0;

	/* configure the ep93xx ide controller for the dma operation. */
	if (rq_data_dir(HWGROUP(drive)->rq) == READ)
		ep93xx_rwproc(drive, 0);
	else
		ep93xx_rwproc(drive, 1);

	/* start the dma transfer. */
	ep93xx_dma_start(hwif->hw.dma, 1, NULL);
	
	DPRINTK("%s end\n", drive->name);
}

static void ep93xx_set_pio(void)
{
	/* Insure that neither device is selected */
	outl(inl(IDECR) | (IDECtrl_CS0n | IDECtrl_CS1n), IDECR);

	/* Clear the MDMA and UDMA operation registers. */
	outl(0, IDEMDMAOP);
	outl(0, IDEUDMAOP);

	/* Reset the UDMA state machine. */
	outl(IDEUDMADebug_RWOE | IDEUDMADebug_RWPTR | IDEUDMADebug_RWDR |
	     IDEUDMADebug_RROE | IDEUDMADebug_RRPTR | IDEUDMADebug_RRDR,
	     IDEUDMADEBUG);
	outl(0, IDEUDMADEBUG);

	/* Enable PIO mode of operation. */
	outl(IDECfg_PIO | IDECfg_IDEEN | (4 << IDECfg_MODE_SHIFT) |
	     (1 << IDECfg_WST_SHIFT), IDECFG);
}

/*
 * ep93xx_ide_dma_end() - cleans up after a dma transfer and checks for errors
 *
 * This checks for the DMA read and write errors that can occur with
 * ep93xx rev. e0.
 */
static int ep93xx_ide_dma_end(ide_drive_t *drive)
{
	ide_hwif_t *hwif = HWIF(drive);
	int i = 0;

	DPRINTK("%s\n", drive->name);

	/*
	 * ep93xx rev e0 error detection.
	 * See if there is any data left in UDMA FIFOs.  For a read, first wait
	 * until either the DMA is done or the FIFO is empty.  If there is any
	 * residual data in the FIFO, there was an error in the transfer.
	 */
	if (rq_data_dir(HWGROUP(drive)->rq) == READ) {
		do {
			i = inl(IDEUDMARFST);
		} while (!ep93xx_dma_is_done(hwif->hw.dma) &&
			 ((i & 15) != ((i >> 4) & 15)));
		udelay(1);
	} else
		i = inl(IDEUDMAWFST);

	if ((i & 15) != ((i >> 4) & 15)) {
		printk(KERN_WARNING "%s: DMA transfer error detected: "
			"fifo residual data\n", drive->name);
		g_bad = 1;
	}

	/*
	 * ep93xx rev e0 error detection.
	 * See if all of the buffers were returned by the DMA driver.  If not,
	 * then a transfer error has occurred.
	 */
	if (!ep93xx_dma_is_done(hwif->hw.dma)) {
		printk(KERN_WARNING "%s: DMA transfer error detected: "
			"transfer not complete\n", drive->name);
		g_bad = 1;
	}

	ep93xx_dma_pause(hwif->hw.dma, 1, 0);
	ep93xx_dma_flush(hwif->hw.dma);

	ep93xx_set_pio();

	/* Indicate there's no dma transfer currently in progress. */
	drive->waiting_for_dma = 0;

	/* compute the drive array index used for this request. */
	i = rq_data_dir(HWGROUP(drive)->rq) == READ ? 0 : 1;
	i += drive->name[2] == 'a' ? 0 : 2;

	if (g_bad == 0) {
		/* reset the consecutive error counter. */
		g_errors[i] = 0;		
	} else {
		g_errors[i]++;
		if (g_errors[i] == FAILURES_BEFORE_DMA_BACKOFF) {

			printk(KERN_WARNING "%s: lowering UDMA mode\n",
				drive->name);

			switch (drive->current_speed) {
			case XFER_UDMA_4:
				ide_config_drive_speed(drive, XFER_UDMA_3);
				break;
			case XFER_UDMA_3:
				ide_config_drive_speed(drive, XFER_UDMA_2);
				break;
			case XFER_UDMA_2:
				ide_config_drive_speed(drive, XFER_UDMA_1);
				break;
			case XFER_UDMA_1:
				ide_config_drive_speed(drive, XFER_UDMA_0);
				break;
			case XFER_UDMA_0:
				HWIF(drive)->ide_dma_off_quietly(drive);
				ide_config_drive_speed(drive, XFER_PIO_4);
				break;
			}

			/* reset the consecutive error counters */
			g_errors[i] = 0;
			g_errors[i ^ 1] = 0;

		}
		return 1;
	}

	return 0;
}

/*
 * ep93xx_ide_dma_test_irq() - checks if the IDE interrupt is asserted 
 */
static int ep93xx_ide_dma_test_irq(ide_drive_t *drive)
{
	if (!drive->waiting_for_dma)
		printk(KERN_WARNING "%s called while not waiting\n",
			drive->name);

	return inl(IDECR) & IDECtrl_INTRQ;
}

static int ep93xx_ide_dma_timeout(ide_drive_t *drive)
{
	printk(KERN_ERR "%s: timeout waiting for DMA\n", drive->name);

	if (HWIF(drive)->ide_dma_test_irq(drive))
		return 0;

	return HWIF(drive)->ide_dma_end(drive);
}

static int ep93xx_ide_dma_lostirq(ide_drive_t *drive)
{
	printk("%s: DMA interrupt recovery\n", drive->name);
	return 1;
}

void ep93xx_ide_dma_init(ide_hwif_t *hwif)
{
	unsigned long flags;

	local_irq_save(flags);

	if (((inl(SYSCON_CHIPID) & SYSCON_CHIPID_REV_MASK)
	     >> SYSCON_CHIPID_REV_SHIFT) <= 0x3) {
		printk(KERN_NOTICE "ep93xx chip revision too old for IDE "
			"DMA support\n");
		goto init_end;
	}

	g_dma_table_base = kmalloc(PRD_ENTRIES * PRD_BYTES, GFP_KERNEL);
	if (g_dma_table_base == NULL) {
		printk("%s: DMA disabled, unable to allocate memory.\n",
		       hwif->name);
		goto init_end;
	}

	/*
	 * Init the dma handle to 0.  This field is used to hold a
	 * handle to the dma instance.
	 */
	hwif->hw.dma = 0;

 	DPRINTK("\n hwif name : %s \n \n", hwif->name);
 
	/* open a ep93xx dma channel instance for ide */
	if (ep93xx_dma_request(&hwif->hw.dma, hwif->name, DMA_IDE) != 0) {
		kfree(g_dma_table_base);
		printk("%s: SG-DMA disabled, unable to allocate DMA channel.\n",
		       hwif->name);
		goto init_end;
	}

	DPRINTK("\n ide init- dma channel allocated: 0x%x, %s  %d \n", 
		hwif->hw.dma, hwif->name, DMA_IDE);

	hwif->atapi_dma			= 1;

	hwif->ultra_mask		= 0x0f;
	hwif->mwdma_mask		= 0x00;
	hwif->swdma_mask		= 0x00;

	hwif->speedproc			= NULL;
	hwif->autodma			= 1;
   
	hwif->ide_dma_check		= ep93xx_ide_dma_check;
	hwif->ide_dma_host_off		= ep93xx_ide_dma_host_off;
	hwif->ide_dma_off_quietly	= ep93xx_ide_dma_off_quietly;
	hwif->ide_dma_host_on		= ep93xx_ide_dma_host_on;
	hwif->ide_dma_on		= ep93xx_ide_dma_on;
	hwif->dma_setup			= ep93xx_ide_dma_setup;
	hwif->dma_exec_cmd		= ep93xx_ide_dma_exec_cmd;
	hwif->dma_start			= ep93xx_ide_dma_start;
	hwif->ide_dma_end		= ep93xx_ide_dma_end;
	hwif->ide_dma_test_irq		= ep93xx_ide_dma_test_irq;
	hwif->ide_dma_timeout		= ep93xx_ide_dma_timeout;
	hwif->ide_dma_lostirq		= ep93xx_ide_dma_lostirq;

 init_end:
	local_irq_restore(flags);
}
EXPORT_SYMBOL(ep93xx_ide_dma_init);

