/*
 *  drivers/mtd/nand/armadillo2x0.c
 *
 *  Copyright (C) 2006,2008 Atmark Techno, Inc.
 */

#include <linux/slab.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/sizes.h>
#include <linux/delay.h>

#define NAND_CMD_CTRL_IRQ_SAVE

//#define NAND_HWCONTROL_PRI_NDELAY  1
//#define NAND_HWCONTROL_POST_NDELAY 1

/*
 * MTD structure for Armadillo-2x0
 */
static struct mtd_info *armadillo2x0_mtd = NULL;

/*
 * Values specific to the Armadillo-2x0
 */
#define ARMADILLO2X0_FIO_PBASE 0x30000000	/* Phys address of flash */
#define ARMADILLO2X0_PXDR	GPIO_PDDR	/*
				 * IO offset to Port D data register
				 * where the CLE, ALE and NCE pins
				 * are wired to.
				 */
#define ARMADILLO2X0_PXDDR	GPIO_PDDDR	/*
				 * IO offset to Port D data direction
				 * register so we can control the IO
				 * lines.
				 */

/*
 * Module stuff
 */

static unsigned long armadillo2x0_fio_pbase = ARMADILLO2X0_FIO_PBASE;
static void __iomem * armadillo2x0_pxdr = (void __iomem *) ARMADILLO2X0_PXDR;
static void __iomem * armadillo2x0_pxddr = (void __iomem *) ARMADILLO2X0_PXDDR;

#ifdef CONFIG_MTD_PARTITIONS
/*
 * Define static partitions for flash device
 */
static struct mtd_partition partition_info[] = {
	{
		.name = "Armadillo-2x0 Nand Flash",
		.offset = 0,
		.size = MTDPART_SIZ_FULL
	},
};
#define NUM_PARTITIONS 1

#endif


/* 
 *	hardware specific access to control-lines
 */
static void armadillo2x0_hwcontrol(struct mtd_info *mtd, int cmd, unsigned int ctrl) 
{
	struct nand_chip *chip = mtd->priv;
#ifdef NAND_CMD_CTRL_IRQ_SAVE
	unsigned long flags;

	local_irq_save(flags);
#endif

#ifdef NAND_HWCONTROL_PRI_NDELAY
	ndelay(NAND_HWCONTROL_PRI_NDELAY);
#endif

	if (ctrl & NAND_CTRL_CHANGE)
		__raw_writel((__raw_readl(armadillo2x0_pxdr) & ~0x70) |
			     ((ctrl & NAND_CLE) ? 0x20 : 0x00) |
			     ((ctrl & NAND_ALE) ? 0x40 : 0x00) |
			     ((ctrl & NAND_NCE) ? 0x00 : 0x10),
			     armadillo2x0_pxdr);

#ifdef NAND_HWCONTROL_POST_NDELAY
	ndelay(NAND_HWCONTROL_POST_NDELAY);
#endif

	if (cmd != NAND_CMD_NONE)
		writeb(cmd, chip->IO_ADDR_W);

#ifdef NAND_CMD_CTRL_IRQ_SAVE
	local_irq_restore(flags);
#endif
}

#ifdef CONFIG_MTD_PARTITIONS
const char *part_probes[] = { "cmdlinepart", NULL };
#endif

/*
 * Main initialization routine
 */
static int __init armadillo2x0_init (void)
{
	struct nand_chip *this;
	const char *part_type = 0;
	int mtd_parts_nb = 0;
	struct mtd_partition *mtd_parts = 0;
	void __iomem * armadillo2x0_fio_base;
	unsigned long temp;

	/* Allocate memory for MTD device structure and private data */
	armadillo2x0_mtd = kmalloc(sizeof(struct mtd_info) + 
			     sizeof(struct nand_chip),
			     GFP_KERNEL);
	if (!armadillo2x0_mtd) {
		printk("Unable to allocate Armadillo-2x0 NAND MTD device structure.\n");
		return -ENOMEM;
	}

	/* map physical adress */
	armadillo2x0_fio_base = ioremap(armadillo2x0_fio_pbase, SZ_1K);
	if(!armadillo2x0_fio_base) {
		printk("ioremap Armadillo-2x0 NAND flash failed\n");
		kfree(armadillo2x0_mtd);
		return -EIO;
	}

	/* Get pointer to private data */
	this = (struct nand_chip *) (&armadillo2x0_mtd[1]);

	/* Initialize structures */
	memset((char *) armadillo2x0_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	armadillo2x0_mtd->priv = this;
	armadillo2x0_mtd->owner = THIS_MODULE;

	temp = inl(SYSCON_DEVCFG);
	outl(0xaa, SYSCON_SWLOCK);
	outl(temp | 0x00000002, SYSCON_DEVCFG);
	outl(0x000004c9, SMCBCR3);

	outl(0x00000010, armadillo2x0_pxdr);
	outl(0x00000070, armadillo2x0_pxddr);

	/* insert callbacks */
	this->IO_ADDR_R = armadillo2x0_fio_base;
	this->IO_ADDR_W = armadillo2x0_fio_base;
	this->cmd_ctrl = armadillo2x0_hwcontrol;
	this->dev_ready = NULL;
	/* 25 us command delay time */
	this->chip_delay = 25;
	this->ecc.mode = NAND_ECC_SOFT;

	/* Scan to find existence of the device */
	if (nand_scan (armadillo2x0_mtd, 1)) {
		iounmap((void *)armadillo2x0_fio_base);
		kfree (armadillo2x0_mtd);
		return -ENXIO;
	}

#ifdef CONFIG_MTD_PARTITIONS
	armadillo2x0_mtd->name = "armadillo2x0-nand";
	mtd_parts_nb = parse_mtd_partitions(armadillo2x0_mtd, part_probes,
					    &mtd_parts, 0);
	if (mtd_parts_nb > 0)
		part_type = "command line";
	else
		mtd_parts_nb = 0;
#endif
	if (mtd_parts_nb == 0) {
		mtd_parts = partition_info;
		mtd_parts_nb = NUM_PARTITIONS;
		part_type = "static";
	}

	/* Register the partitions */
	printk(KERN_NOTICE "Using %s partition definition\n", part_type);
	add_mtd_partitions(armadillo2x0_mtd, mtd_parts, mtd_parts_nb);

	/* Return happy */
	return 0;
}
module_init(armadillo2x0_init);

/*
 * Clean up routine
 */
static void __exit armadillo2x0_cleanup (void)
{
	struct nand_chip *this = (struct nand_chip *)&armadillo2x0_mtd[1];

	/* Release resources, unregister device */
	nand_release (armadillo2x0_mtd);

	/* Release iomaps */
	iounmap(this->IO_ADDR_R);

	/* Free the MTD device structure */
	kfree (armadillo2x0_mtd);
}
module_exit(armadillo2x0_cleanup);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Atmark Techno, Inc.");
MODULE_DESCRIPTION("MTD map driver for Cogent Armadillo-2x0");
